# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable

import jax
import jax.numpy as jnp
import numpy as np

__all__ = ["fixpoint_iter", "fixpoint_iter_backprop"]


def fixpoint_iter(
    cond_fn: Callable[[int, Any, Any], bool],
    body_fn: Callable[[Any, Any, Any, Any], Any], min_iterations: int,
    max_iterations: int, inner_iterations: int, constants: Any, state: Any
):
  """Implementation of a fixed point loop.

  This fixed point loop iterator applies ``body_fn`` to a tuple
  ``(iteration, constants, state, compute_error)`` to output a new state, using
  context provided in iteration and constants.

  ``body_fn`` is iterated (inner_iterations -1) times, and one last time with
  the ``compute_error`` flag to ``True``, indicating that additional
  computational effort can be spent on recalculating the latest error
  (``errors`` are stored as the first element of the state tuple).

  upon termination of these ``inner_iterations``, the loop is continued if
  iteration is smaller than ``min_iterations``, stopped if equal/larger than
  ``max_iterations``, and interrupted if ``cond_fn`` returns False.

  Args:
    cond_fn : termination condition function
    body_fn : body loop instructions
    min_iterations : lower bound on the total amount of fixed point iterations
    max_iterations : upper bound on the total amount of fixed point iterations
    inner_iterations : number of iterations ``body_fn`` will be executed
      successively before calling ``cond_fn``.
    constants : constant (during loop) parameters passed on to body
    state : state variable

  Returns:
    outputs state returned by ``body_fn`` upon termination.
  """  # noqa: D401
  # If number of minimal iterations matches maximal number, force a scan instead
  # of a while loop.

  force_scan = (min_iterations == max_iterations)

  compute_error_flags = jnp.arange(inner_iterations) == inner_iterations - 1

  def max_cond_fn(iteration_state):
    iteration, state = iteration_state
    return jnp.logical_and(
        iteration < max_iterations,
        jnp.logical_or(
            iteration < min_iterations, cond_fn(iteration, constants, state)
        )
    )

  def unrolled_body_fn(iteration_state):

    def one_iteration(iteration_state, compute_error):
      iteration, state = iteration_state
      state = body_fn(iteration, constants, state, compute_error)
      iteration += 1
      return (iteration, state), None

    iteration_state, _ = jax.lax.scan(
        one_iteration, iteration_state, compute_error_flags
    )
    return (iteration_state, None) if force_scan else iteration_state

  if force_scan:
    (_, state), _ = jax.lax.scan(
        lambda carry, x: unrolled_body_fn(carry), (0, state),
        None,
        length=max_iterations // inner_iterations
    )
  else:
    _, state = jax.lax.while_loop(max_cond_fn, unrolled_body_fn, (0, state))
  return state


def fixpoint_iter_fwd(
    cond_fn, body_fn, min_iterations, max_iterations, inner_iterations,
    constants, state
):
  """Forward iteration of fixed point iteration to handle backpropagation.

  The main difference with fixpoint_iter is the checkpointing, in variable
  states, of the state variables as they are recorded through iterations, every
  inner_iterations. This sequence of states will be used in the backward loop.

  Args:
    cond_fn : termination condition function
    body_fn : body loop instructions
    min_iterations : lower bound on the total amount of fixed point iterations
    max_iterations : upper bound on the total amount of fixed point iterations
    inner_iterations : number of iterations body_fn will be executed
      successively before calling cond_fn.
    constants : constant (during loop) parameters passed on to body
    state : state variable

  Returns:
    outputs state returned by body_fn upon termination.
  """
  force_scan = min_iterations == max_iterations
  compute_error_flags = jnp.arange(inner_iterations) == inner_iterations - 1
  states = jax.tree_util.tree_map(
      lambda x: jnp.zeros(
          (max_iterations // inner_iterations + 1,) + jnp.shape(x),
          dtype=jax.dtypes.result_type(x)
      ), state
  )

  def max_cond_fn(iteration_states_state):
    iteration, _, state = iteration_states_state
    return jnp.logical_and(
        iteration < max_iterations,
        jnp.logical_or(
            iteration < min_iterations, cond_fn(iteration, constants, state)
        )
    )

  def unrolled_body_fn(iteration_states_state):
    iteration, states, state = iteration_states_state
    states = jax.tree_util.tree_map(
        lambda states, state: jax.lax.dynamic_update_index_in_dim(
            states, state, iteration // inner_iterations, 0
        ), states, state
    )

    def one_iteration(iteration_state, compute_error):
      iteration, state = iteration_state
      state = body_fn(iteration, constants, state, compute_error)
      iteration += 1
      return (iteration, state), None

    iteration_state, _ = jax.lax.scan(
        one_iteration, (iteration, state), compute_error_flags
    )
    iteration, state = iteration_state
    out = (iteration, states, state)
    return (out, None) if force_scan else out

  if force_scan:
    (iteration, states, state), _ = jax.lax.scan(
        lambda carry, x: unrolled_body_fn(carry), (0, states, state),
        None,
        length=max_iterations // inner_iterations
    )
  else:
    iteration, states, state = jax.lax.while_loop(
        max_cond_fn, unrolled_body_fn, (0, states, state)
    )

  return state, (constants, iteration, states)


def fixpoint_iter_bwd(
    cond_fn, body_fn, min_iterations, max_iterations, inner_iterations, res, g
):
  """Backward iteration of fixed point iteration, using checkpointed states."""
  del cond_fn
  force_scan = (min_iterations == max_iterations)
  constants, iteration, states = res
  # The tree may contain some python floats
  g_constants = jax.tree_util.tree_map(
      lambda x: jnp.zeros_like(x, dtype=x.dtype)
      if isinstance(x, (np.ndarray, jnp.ndarray)) else 0, constants
  )

  def bwd_cond_fn(iteration_g_gconst):
    iteration, _, _ = iteration_g_gconst
    return iteration >= 0

  def unrolled_body_fn_no_errors(iteration, constants, state):
    compute_error_flags = jnp.zeros((inner_iterations,), dtype=bool)

    def one_iteration(iteration_state, compute_error):
      iteration, state = iteration_state
      state = body_fn(iteration, constants, state, compute_error)
      iteration += 1
      return (iteration, state), None

    iteration_state, _ = jax.lax.scan(
        one_iteration, (iteration, state), compute_error_flags
    )
    _, state = iteration_state
    return state

  def unrolled_body_fn(iteration_g_gconst):
    iteration, g, g_constants = iteration_g_gconst
    state = jax.tree_util.tree_map(
        lambda x: x[iteration // inner_iterations], states
    )
    _, pullback = jax.vjp(
        unrolled_body_fn_no_errors, iteration, constants, state
    )
    _, gi_constants, g_state = pullback(g)
    g_constants = jax.tree_util.tree_map(
        lambda x, y: x + y, g_constants, gi_constants
    )
    out = (iteration - inner_iterations, g_state, g_constants)
    return (out, None) if force_scan else out

  if force_scan:
    (_, g_state, g_constants), _ = jax.lax.scan(
        lambda carry, x: unrolled_body_fn(carry), (0, g, g_constants),
        None,
        length=max_iterations // inner_iterations
    )
  else:
    _, g_state, g_constants = jax.lax.while_loop(
        bwd_cond_fn, unrolled_body_fn,
        (iteration - inner_iterations, g, g_constants)
    )

  return g_constants, g_state


# definition of backprop friendly variant of fixpoint_iter.
fixpoint_iter_backprop = jax.custom_vjp(
    fixpoint_iter, nondiff_argnums=(0, 1, 2, 3, 4)
)

fixpoint_iter_backprop.defvjp(fixpoint_iter_fwd, fixpoint_iter_bwd)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import TYPE_CHECKING, Any, Optional

from ott.geometry import pointcloud
from ott.problems.quadratic import quadratic_problem
from ott.solvers import linear
from ott.solvers.linear import sinkhorn

if TYPE_CHECKING:
  from ott.geometry import distrib_costs

__all__ = ["third_lower_bound"]


def third_lower_bound(
    prob: quadratic_problem.QuadraticProblem,
    distrib_cost: "distrib_costs.UnivariateWasserstein",
    epsilon: Optional[float] = None,
    **kwargs: Any,
) -> sinkhorn.SinkhornOutput:
  """Computes the third lower bound distance from :cite:`memoli:11`, def. 6.3.

  Args:
    prob: Quadratic OT problem.
    distrib_cost: Univariate Wasserstein cost used to compare two point clouds
      in different spaces. Each point is seen as its distribution of costs
      to other points in its respective point cloud.
    epsilon: Entropy regularization.
    kwargs: Keyword arguments for :func:`~ott.solvers.linear.solve`.

  Returns:
    An approximation of the GW coupling that can be used to initialize
    the solution of the quadratic OT problem.
  """
  dists_xx = prob.geom_xx.cost_matrix
  dists_yy = prob.geom_yy.cost_matrix
  geom_xy = pointcloud.PointCloud(
      dists_xx, dists_yy, cost_fn=distrib_cost, epsilon=epsilon
  )

  return linear.solve(geom_xy, **kwargs)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
from typing import TYPE_CHECKING, Any, Dict, Optional, Sequence, Tuple

import jax
import jax.numpy as jnp

from ott.geometry import geometry

if TYPE_CHECKING:
  from ott.problems.linear import linear_problem
  from ott.problems.quadratic import quadratic_problem

__all__ = ["BaseQuadraticInitializer", "QuadraticInitializer"]


@jax.tree_util.register_pytree_node_class
class BaseQuadraticInitializer(abc.ABC):
  """Base class for quadratic initializers.

  Args:
    kwargs: Keyword arguments.
  """

  def __init__(self, **kwargs: Any):
    self._kwargs = kwargs

  def __call__(
      self, quad_prob: "quadratic_problem.QuadraticProblem", **kwargs: Any
  ) -> "linear_problem.LinearProblem":
    """Compute the initial linearization of a quadratic problem.

    Args:
      quad_prob: Quadratic problem to linearize.
      kwargs: Additional keyword arguments.

    Returns:
      Linear problem.
    """
    from ott.problems.linear import linear_problem

    n, m = quad_prob.geom_xx.shape[0], quad_prob.geom_yy.shape[0]
    geom = self._create_geometry(quad_prob, **kwargs)
    assert geom.shape == (n, m), (
        f"Expected geometry of shape `{n, m}`, "
        f"found `{geom.shape}`."
    )
    return linear_problem.LinearProblem(
        geom,
        a=quad_prob.a,
        b=quad_prob.b,
        tau_a=quad_prob.tau_a,
        tau_b=quad_prob.tau_b,
    )

  @abc.abstractmethod
  def _create_geometry(
      self, quad_prob: "quadratic_problem.QuadraticProblem", **kwargs: Any
  ) -> geometry.Geometry:
    """Compute initial geometry for linearization.

    Args:
      quad_prob: Quadratic problem.
      kwargs: Additional keyword arguments.

    Returns:
      Geometry used to initialize the linearized problem.
    """

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:  # noqa: D102
    return [], self._kwargs

  @classmethod
  def tree_unflatten(  # noqa: D102
      cls, aux_data: Dict[str, Any], children: Sequence[Any]
  ) -> "BaseQuadraticInitializer":
    return cls(*children, **aux_data)


class QuadraticInitializer(BaseQuadraticInitializer):
  r"""Initialize a linear problem locally around a selected coupling.

  If the problem is balanced (``tau_a = 1`` and ``tau_b = 1``),
  the equation of the cost follows eq. 6, p. 1 of :cite:`peyre:16`.

  If the problem is unbalanced (``tau_a < 1`` or ``tau_b < 1``), there are two
  possible cases. A first possibility is to introduce a quadratic KL
  divergence on the marginals in the objective as done in :cite:`sejourne:21`
  (``gw_unbalanced_correction = True``), which in turns modifies the
  local cost matrix.

  Alternatively, it could be possible to leave the formulation of the
  local cost unchanged, i.e. follow eq. 6, p. 1 of :cite:`peyre:16`
  (``gw_unbalanced_correction = False``) and include the unbalanced terms
  at the level of the linear problem only.

  Let :math:`P` [num_a, num_b] be the transport matrix, `cost_xx` is the
  cost matrix of `geom_xx` and `cost_yy` is the cost matrix of `geom_yy`.
  `left_x` and `right_y` depend on the loss chosen for GW.
  `gw_unbalanced_correction` is flag indicating whether the unbalanced
  correction applies. The equation of the local cost can be written as:

  .. math::

    \text{marginal_dep_term} + \text{left}_x(\text{cost_xx}) P
    \text{right}_y(\text{cost_yy}) + \text{unbalanced_correction}

  When working with the fused problem, a linear term is added to the cost
  matrix: `cost_matrix` += `fused_penalty` * `geom_xy.cost_matrix`

  Args:
    init_coupling: The coupling to use for initialization. If :obj:`None`,
      defaults to the product coupling :math:`ab^T`.
  """

  def __init__(
      self, init_coupling: Optional[jnp.ndarray] = None, **kwargs: Any
  ):
    super().__init__(**kwargs)
    self.init_coupling = init_coupling

  def _create_geometry(
      self,
      quad_prob: "quadratic_problem.QuadraticProblem",
      *,
      epsilon: float,
      relative_epsilon: Optional[bool] = None,
      **kwargs: Any,
  ) -> geometry.Geometry:
    """Compute initial geometry for linearization.

    Args:
      quad_prob: Quadratic OT problem.
      epsilon: Epsilon regularization.
      relative_epsilon: Flag, use `relative_epsilon` or not in geometry.
      kwargs: Keyword arguments for :class:`~ott.geometry.geometry.Geometry`.

    Returns:
      The initial geometry used to initialize the linearized problem.
    """
    from ott.problems.quadratic import quadratic_problem

    del kwargs

    marginal_cost = quad_prob.marginal_dependent_cost(quad_prob.a, quad_prob.b)
    geom_xx, geom_yy = quad_prob.geom_xx, quad_prob.geom_yy

    h1, h2 = quad_prob.quad_loss
    if self.init_coupling is None:
      tmp1 = quadratic_problem.apply_cost(geom_xx, quad_prob.a, axis=1, fn=h1)
      tmp2 = quadratic_problem.apply_cost(geom_yy, quad_prob.b, axis=1, fn=h2)
      tmp = jnp.outer(tmp1, tmp2)
    else:
      tmp1 = h1.func(geom_xx.cost_matrix)
      tmp2 = h2.func(geom_yy.cost_matrix)
      tmp = tmp1 @ self.init_coupling @ tmp2.T

    if quad_prob.is_balanced:
      cost_matrix = marginal_cost.cost_matrix - tmp
    else:
      # initialize epsilon for Unbalanced GW according to Sejourne et. al (2021)
      init_transport = jnp.outer(quad_prob.a, quad_prob.b)
      marginal_1, marginal_2 = init_transport.sum(1), init_transport.sum(0)

      epsilon *= marginal_1.sum()
      unbalanced_correction = quad_prob.cost_unbalanced_correction(
          init_transport, marginal_1, marginal_2, epsilon=epsilon
      )
      cost_matrix = marginal_cost.cost_matrix - tmp + unbalanced_correction

    cost_matrix += quad_prob.fused_penalty * quad_prob._fused_cost_matrix
    return geometry.Geometry(
        cost_matrix=cost_matrix,
        epsilon=epsilon,
        relative_epsilon=relative_epsilon
    )

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:  # noqa: D102
    return [self.init_coupling], self._kwargs



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
from typing import Callable, Literal, NamedTuple, Optional

import jax.numpy as jnp
from jaxopt import LBFGS

from ott import utils

__all__ = [
    "ConjugateResults",
    "FenchelConjugateSolver",
    "FenchelConjugateLBFGS",
    "DEFAULT_CONJUGATE_SOLVER",
]


class ConjugateResults(NamedTuple):
  r"""Holds the results of numerically conjugating a function.

  Args:
    val: the conjugate value, i.e., :math:`f^\star(y)`
    grad: the gradient, i.e., :math:`\nabla f^\star(y)`
    num_iter: the number of iterations taken by the solver
  """
  val: float
  grad: jnp.ndarray
  num_iter: int


class FenchelConjugateSolver(abc.ABC):
  r"""Abstract conjugate solver class.

  Given a function :math:`f`, numerically estimate the Fenchel conjugate
  :math:`f^\star(y) := -\inf_{x\in\mathbb{R}^n} f(x)-\langle x, y\rangle`.
  """

  @abc.abstractmethod
  def solve(
      self,
      f: Callable[[jnp.ndarray], jnp.ndarray],
      y: jnp.ndarray,
      x_init: Optional[jnp.ndarray] = None
  ) -> ConjugateResults:
    """Solve for the conjugate.

    Args:
      f: function to conjugate
      y: point to conjugate
      x_init: initial point to search over

    Returns:
      The solution to the conjugation.
    """


@utils.register_pytree_node
class FenchelConjugateLBFGS(FenchelConjugateSolver):
  """Solve for the conjugate using :class:`~jaxopt.LBFGS`.

  Args:
    gtol: gradient tolerance
    max_iter: maximum number of iterations
    max_linesearch_iter: maximum number of line search iterations
    linesearch_type: type of line search
    linesearch_init: strategy for line search initialization
    increase_factor: factor by which to increase the step size during
      the line search
  """

  gtol: float = 1e-3
  max_iter: int = 10
  max_linesearch_iter: int = 10
  linesearch_type: Literal["zoom", "backtracking",
                           "hager-zhang"] = "backtracking"
  linesearch_init: Literal["increase", "max", "current"] = "increase"
  increase_factor: float = 1.5

  def solve(  # noqa: D102
      self,
      f: Callable[[jnp.ndarray], jnp.ndarray],
      y: jnp.ndarray,
      x_init: Optional[jnp.array] = None
  ) -> ConjugateResults:
    assert y.ndim == 1, y.ndim

    solver = LBFGS(
        fun=lambda x: f(x) - x.dot(y),
        tol=self.gtol,
        maxiter=self.max_iter,
        linesearch=self.linesearch_type,
        linesearch_init=self.linesearch_init,
        increase_factor=self.increase_factor,
        implicit_diff=False,
        unroll=False
    )

    out = solver.run(y if x_init is None else x_init)
    return ConjugateResults(
        val=-out.state.value, grad=out.params, num_iter=out.state.iter_num
    )


DEFAULT_CONJUGATE_SOLVER = FenchelConjugateLBFGS(
    gtol=1e-5,
    max_iter=20,
    max_linesearch_iter=20,
    linesearch_type="backtracking",
)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import jax.numpy as jnp

__all__ = ["cyclical_time_encoder"]


def cyclical_time_encoder(t: jnp.ndarray, n_freqs: int = 128) -> jnp.ndarray:
  r"""Encode time :math:`t` into a cyclical representation.

  Time :math:`t` is encoded as :math:`cos(\hat{t})` and :math:`sin(\hat{t})`
  where :math:`\hat{t} = [2\pi t, 2\pi 2 t,\dots, 2\pi n_f t]`.

  Args:
    t: Time of shape ``[n, 1]``.
    n_freqs: Frequency :math:`n_f` of the cyclical encoding.

  Returns:
    Encoded time of shape ``[n, 2 * n_freqs]``.
  """
  freq = 2 * jnp.arange(n_freqs) * jnp.pi
  t = freq * t
  return jnp.concatenate([jnp.cos(t), jnp.sin(t)], axis=-1)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Protocol

import jax.numpy as jnp

__all__ = ["Transport"]

# TODO(michalk8): introduce additional types here


class Transport(Protocol):
  """Interface for the solution of a transport problem.

  Classes implementing those function do not have to inherit from it, the
  class can however be used in type hints to support duck typing.
  """

  @property
  def matrix(self) -> jnp.ndarray:
    ...

  def apply(self, inputs: jnp.ndarray, axis: int) -> jnp.ndarray:
    ...

  def marginal(self, axis: int = 0) -> jnp.ndarray:
    ...



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, NamedTuple, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np

from ott.geometry import costs, epsilon_scheduler, pointcloud
from ott.math import fixed_point_loop
from ott.math import utils as mu

__all__ = ["MMSinkhornOutput", "MMSinkhorn"]


class MMSinkhornState(NamedTuple):
  potentials: Tuple[jnp.ndarray, ...]
  errors: jnp.ndarray

  def solution_error(
      self,
      cost_t: jnp.ndarray,
      a_s: Tuple[jnp.ndarray, ...],
      epsilon: float,
      norm_error: float = 1.0
  ) -> float:
    coupl_tensor = coupling_tensor(self.potentials, cost_t, epsilon)
    marginals = tensor_marginals(coupl_tensor)
    errors = jnp.array([
        jnp.sum(jnp.abs(a - marginal) ** norm_error) ** (1.0 / norm_error)
        for a, marginal in zip(a_s, marginals)
    ])
    return jnp.sum(errors)

  def set(self, **kwargs: Any) -> "MMSinkhornState":
    """Return a copy of self, with potential overwrites."""
    return self._replace(**kwargs)


class MMSinkhornOutput(NamedTuple):
  r"""Output of the MMSinkhorn solver used on :math:`k` point clouds.

  This class contains both solutions and problem definition of a regularized
  MM-OT problem involving :math:`k` weighted point clouds of varying sizes,
  along with methods and properties that can use or describe the solution.

  Args:
    potentials: Tuple of :math:`k` optimal dual variables, vectors of sizes
      equal to the number of points in each of the :math:`k` point clouds.
    errors: Vector of errors, along iterations. This vector is of size
      ``max_iterations // inner_iterations`` where those were the parameters
      passed on to the :class:`~ott.experimental.mmsinkhorn.MMSinkhorn` solver.
      Follows the conventions used in
      :attr:`~ott.solvers.linear.sinkhorn.SinkhornOutput.errors`
    x_s: Tuple of :math:`k` point clouds, ``x_s[i]`` is a matrix of size
      :math:`n_i \times d` where `d` is common to all point clouds.
    a_s: Tuple of :math:`k` probability vectors, each of size :math:`n_i`.
    cost_fns: Cost function, or a tuple of :math:`k(k-1)/2` such instances.
    epsilon: Entropic regularization used to solve the multimarginal Sinkhorn
      problem.
    ent_reg_cost: The regularized optimal transport cost, the linear
      contribution (dot product between optimal tensor and cost) minus entropy
      times ``epsilon``.
    threshold: Convergence threshold used to control the termination of the
      algorithm.
    converged: Whether the output corresponds to a solution whose error is
      below the convergence threshold.
    inner_iterations: Number of iterations that were run between two
      computations of errors.
  """
  potentials: Tuple[jnp.ndarray, ...]
  errors: jnp.ndarray
  x_s: Optional[Tuple[jnp.ndarray, ...]] = None
  a_s: Optional[Tuple[jnp.ndarray, ...]] = None
  cost_fns: Optional[Union[costs.CostFn, Tuple[costs.CostFn, ...]]] = None
  epsilon: Optional[float] = None
  ent_reg_cost: Optional[jnp.ndarray] = None
  threshold: Optional[jnp.ndarray] = None
  converged: Optional[bool] = None
  inner_iterations: Optional[int] = None

  def set(self, **kwargs: Any) -> "MMSinkhornOutput":
    """Return a copy of self, with potential overwrites."""
    return self._replace(**kwargs)

  @property
  def n_iters(self) -> int:  # noqa: D102
    """Total number of iterations that were needed to terminate."""
    return jnp.sum(self.errors != -1) * self.inner_iterations

  @property
  def cost_t(self) -> jnp.ndarray:
    """Cost tensor."""
    return cost_tensor(self.x_s, self.cost_fns)

  @property
  def tensor(self) -> jnp.ndarray:
    """Transport tensor."""
    return jnp.exp(
        -remove_tensor_sum(self.cost_t, self.potentials) / self.epsilon
    )

  @property
  def marginals(self) -> Tuple[jnp.ndarray, ...]:
    """:math:`k` marginal probability weight vectors."""
    return tensor_marginals(self.tensor)

  def marginal(self, k: int) -> jnp.ndarray:
    """Return the marginal probability weight vector at slice :math:`k`."""
    return tensor_marginal(self.tensor, k)

  @property
  def transport_mass(self) -> float:
    """Sum of transport tensor."""
    return jnp.sum(self.tensor)

  @property
  def shape(self) -> Tuple[int, ...]:
    """Shape of the transport :attr:`tensor`."""
    return tuple(x.shape[0] for x in self.x_s)

  @property
  def n_marginals(self) -> int:
    """Number of marginals."""
    return len(self.x_s)


def cost_tensor(
    x_s: Tuple[jnp.ndarray, ...], cost_fns: Union[costs.CostFn,
                                                  Tuple[costs.CostFn, ...]]
) -> jnp.ndarray:
  r"""Create a cost tensor from a tuple of :math:`k` :math:`d`-dim point clouds.

  Args:
    x_s: Tuple of :math:`k` point clouds, each described as a
      :math:`n_i \times d` matrix of batched vectors.
    cost_fns: Either a single :ott:`ott.geometry.costs.CostFn` object, or a
      tuple of :math:`k (k-1)/2` of them. Current implementation only works for
      symmetric and definite cost functions (i.e. such that
      :math:`c(x, y) = c(y, x)` and :math:`c(x, x) = 0`).
  """

  def c_fn_pair(i: int, j: int) -> costs.CostFn:
    if isinstance(cost_fns, costs.CostFn):
      return cost_fns
    return cost_fns[i * k - (i * (i + 1)) // 2 + j - i - 1]

  k = len(x_s)  # TODO(cuturi) padded version
  ns = [x.shape[0] for x in x_s]
  cost_t = jnp.zeros(ns)

  for i in range(k):
    for j in range(i + 1, k):
      cost_m = pointcloud.PointCloud(
          x_s[i], x_s[j], cost_fn=c_fn_pair(i, j)
      ).cost_matrix
      axis = list(range(i)) + list(range(i + 1, j)) + list(range(j + 1, k))
      cost_t += jnp.expand_dims(cost_m, axis=axis)
  return cost_t


def remove_tensor_sum(
    c: jnp.ndarray, u: Tuple[jnp.ndarray, ...]
) -> jnp.ndarray:
  r"""Remove the tensor sum of :math:`k` vectors to tensor of :math:`k` dims.

  Args:
    c: :math:`n_1 \times \cdots n_k` tensor.
    u: Tuple of :math:`k` vectors, each of size :math:`n_i`.

  Return:
    Tensor :math:`c - u[0] \oplus u[1] \oplus ... \oplus u[n]`.
  """
  k = len(u)
  for i in range(k):
    c -= jnp.expand_dims(u[i], axis=list(range(i)) + list(range(i + 1, k)))
  return c


def tensor_marginals(coupling: jnp.ndarray) -> Tuple[jnp.ndarray, ...]:
  return tuple(tensor_marginal(coupling, ix) for ix in range(coupling.ndim))


def tensor_marginal(coupling: jnp.ndarray, slice_index: int) -> jnp.ndarray:
  k = coupling.ndim
  axis = list(range(slice_index)) + list(range(slice_index + 1, k))
  return coupling.sum(axis=axis)


@jtu.register_pytree_node_class
class MMSinkhorn:
  r"""Multimarginal Sinkhorn solver, aligns :math:`k \,d`-dim point clouds.

  This solver implements the entropic multimarginal solver presented in
  :cite:`benamou:15` and described in :cite:`piran:24`, Algorithm 1.
  The current implementation follows largely the template of the
  :class:`~ott.solvers.linear.sinkhorn.Sinkhorn` solver, with a much reduced
  set of hyperparameters, controlling the number of iterations and convergence
  threshold, along with the application of the :cite:`danskin:67` theorem to
  instantiate the OT cost. The iterations are done by default in log-space.

  Args:
    threshold: tolerance used to stop the Sinkhorn iterations. This is
      typically the deviation between a target marginal and the marginal of the
      current primal solution.
    norm_error: power used to define p-norm of error for marginal/target.
    inner_iterations: the Sinkhorn error is not recomputed at each
      iteration but every ``inner_iterations`` instead.
    min_iterations: the minimum number of Sinkhorn iterations carried
      out before the error is computed and monitored.
    max_iterations: the maximum number of Sinkhorn iterations. If
      ``max_iterations`` is equal to ``min_iterations``, Sinkhorn iterations are
      run by default using a :func:`~jax.lax.scan` loop rather than a custom,
      unroll-able :func:`~jax.lax.while_loop` that monitors convergence.
      In that case the error is not monitored and the ``converged``
      flag will return :obj:`False` as a consequence.
    use_danskin: when :obj:`True`, it is assumed the entropy regularized cost
      is evaluated using optimal potentials that are frozen, i.e. whose
      gradients have been stopped. This is useful when carrying out first order
      differentiation, and is only valid mathematically when the algorithm has
      converged with a low tolerance.
  """

  def __init__(
      self,
      threshold: float = 1e-3,
      norm_error: float = 1.0,
      inner_iterations: int = 10,
      min_iterations: int = 0,
      max_iterations: int = 2000,
      use_danskin: bool = True,
  ):
    self.threshold = threshold
    self.inner_iterations = inner_iterations
    self.min_iterations = min_iterations
    self.max_iterations = max_iterations
    self.norm_error = norm_error
    self.use_danskin = use_danskin

  def __call__(
      self,
      x_s: Tuple[jnp.ndarray, ...],
      a_s: Optional[Tuple[jnp.ndarray, ...]] = None,
      cost_fns: Optional[Union[costs.CostFn, Tuple[costs.CostFn, ...]]] = None,
      epsilon: Optional[float] = None
  ) -> MMSinkhornOutput:
    r"""Solve multimarginal OT for :math:`k` :math:`d`-dim point clouds.

    Takes :math:`k` weighted :math:`d`-dim point clouds and computes their
    multimarginal optimal transport tensor. The :math:`d` dimensional point
    clouds are stored in ``x_s``, along with :math:`k` probability vectors,
    stored in ``a_s``, as well as a :class:`~ott.geometry.costs.CostFn`
    instance (or :math:`k(k-1)/2` of them, one for each pair of point clouds
    ``x_s[i]`` and ``x_s[j]``, ``i<j``.)

    The solver also uses ``epsilon`` as an input, with the default rule set to
    one twentieth of the standard deviation of the all values stored in the cost
    tensor resulting from these inputs.

    Args:
      x_s: Tuple of :math:`k` point clouds, ``x_s[i]`` is a matrix of size
        :math:`n_i \times d` where :math:`d` is a dimension common to all
        point clouds.
      a_s: Tuple of :math:`k` probability vectors, each of size :math:`n_i`.
      cost_fns: Instance of :class:`~ott.geometry.costs.CostFn`, or a tuple
        of :math:`k(k-1)/2` such instances. Note that the solver currently
        assumes that these cost functions are symmetric. The cost function at
        index :math:`i(k-\tfrac{i+1}{2})+j-i-1` will be used to compare
        point cloud ``x_s[i]`` with point cloud ``x_s[j]``.
      epsilon: entropic regularization used to solve the multimarginal Sinkhorn
        problem.

    Returns:
      Multimarginal Sinkhorn output.
    """
    n_s = [x.shape[0] for x in x_s]
    if cost_fns is None:
      cost_fns = costs.SqEuclidean()
    elif isinstance(cost_fns, Tuple):
      assert len(cost_fns) == (len(n_s) * (len(n_s) - 1)) // 2

    # Default to uniform probability weights for each point cloud.
    if a_s is None:
      a_s = [jnp.ones(n) / n for n in n_s]
    else:
      # Case in which user passes ``None`` weights within tuple.
      a_s = [(jnp.ones(n) / n if a is None else a) for a, n in zip(a_s, n_s)]

    assert len(n_s) == len(a_s), (len(n_s), len(a_s))
    for n, a in zip(n_s, a_s):
      assert n == a.shape[0], (n, a.shape[0])

    cost_t = cost_tensor(x_s, cost_fns)
    state = self.init_state(n_s)
    if epsilon is None:
      epsilon = epsilon_scheduler.DEFAULT_SCALE * jnp.std(cost_t)
    const = cost_t, a_s, epsilon
    out = run(const, self, state)
    return out.set(x_s=x_s, a_s=a_s, cost_fns=cost_fns, epsilon=epsilon)

  def init_state(self, n_s: Tuple[int, ...]) -> MMSinkhornState:
    """Return the initial state of the loop."""
    errors = -jnp.ones((self.outer_iterations, 1))
    potentials = tuple(jnp.zeros(n) for n in n_s)
    return MMSinkhornState(potentials=potentials, errors=errors)

  def _converged(self, state: MMSinkhornState, iteration: int) -> bool:
    err = state.errors[iteration // self.inner_iterations - 1, 0]
    return jnp.logical_and(iteration > 0, err < self.threshold)

  def _diverged(self, state: MMSinkhornState, iteration: int) -> bool:
    err = state.errors[iteration // self.inner_iterations - 1, 0]
    return jnp.logical_not(jnp.isfinite(err))

  def _continue(self, state: MMSinkhornState, iteration: int) -> bool:
    """Continue while not(converged) and not(diverged)."""
    return jnp.logical_and(
        jnp.logical_not(self._diverged(state, iteration)),
        jnp.logical_not(self._converged(state, iteration))
    )

  @property
  def outer_iterations(self) -> int:
    """Upper bound on number of times inner_iterations are carried out.

    This integer can be used to set constant array sizes to track the algorithm
    progress, notably errors.
    """
    return np.ceil(self.max_iterations / self.inner_iterations).astype(int)

  def tree_flatten(self):  # noqa: D102
    aux = vars(self).copy()
    aux.pop("threshold")
    return [self.threshold], aux

  @classmethod
  def tree_unflatten(cls, aux_data, children):  # noqa: D102
    return cls(**aux_data, threshold=children[0])


def run(
    const: Tuple[jnp.ndarray, Tuple[jnp.ndarray, ...], float],
    solver: MMSinkhorn, state: MMSinkhornState
) -> MMSinkhornOutput:

  def cond_fn(
      iteration: int, const: Tuple[jnp.ndarray, Tuple[jnp.ndarray, ...], float],
      state: MMSinkhornState
  ) -> bool:
    del const
    return solver._continue(state, iteration)

  def body_fn(
      iteration: int, const: Tuple[jnp.ndarray, Tuple[jnp.ndarray, ...], float],
      state: MMSinkhornState, compute_error: bool
  ) -> MMSinkhornState:
    cost_t, a_s, epsilon = const
    k = len(a_s)

    def one_slice(potentials: Tuple[jnp.ndarray, ...], l: int, a: jnp.ndarray):
      pot = potentials[l]
      axis = list(range(l)) + list(range(l + 1, k))
      app_lse = mu.softmin(
          remove_tensor_sum(cost_t, potentials), epsilon, axis=axis
      )
      pot += epsilon * jnp.log(a) + jnp.where(jnp.isfinite(app_lse), app_lse, 0)
      return potentials[:l] + (pot,) + potentials[l + 1:]

    potentials = state.potentials
    for l in range(k):
      potentials = one_slice(potentials, l, a_s[l])

    state = state.set(potentials=potentials)
    err = jax.lax.cond(
        jnp.logical_or(
            iteration == solver.max_iterations - 1,
            jnp.logical_and(compute_error, iteration >= solver.min_iterations)
        ),
        lambda state, c, a, e: state.solution_error(c, a, e, solver.norm_error),
        lambda *_: jnp.inf, state, cost_t, a_s, epsilon
    )
    errors = state.errors.at[iteration // solver.inner_iterations, :].set(err)
    return state.set(errors=errors)

  fix_point = fixed_point_loop.fixpoint_iter_backprop
  state = fix_point(
      cond_fn, body_fn, solver.min_iterations, solver.max_iterations,
      solver.inner_iterations, const, state
  )
  converged = jnp.logical_and(
      jnp.logical_not(jnp.any(jnp.isnan(state.errors))), state.errors[-1, 0]
      < solver.threshold
  )

  out = MMSinkhornOutput(
      potentials=state.potentials,
      errors=state.errors,
      threshold=solver.threshold,
      converged=converged,
      inner_iterations=solver.inner_iterations
  )

  # Compute cost
  if solver.use_danskin:
    potentials = [jax.lax.stop_gradient(pot) for pot in out.potentials]
  else:
    potentials = out.potentials

  cost_t, a_s, epsilon = const
  ent_reg_cost = 0.0
  for potential, a in zip(potentials, a_s):
    pot = jnp.where(jnp.isfinite(potential), potential, 0)
    ent_reg_cost += jnp.sum(pot * a)

  ent_reg_cost += epsilon * (
      1 - jnp.sum(coupling_tensor(potentials, cost_t, epsilon))
  )
  return out.set(ent_reg_cost=ent_reg_cost)


def coupling_tensor(
    potentials: Tuple[jnp.ndarray], cost_t: jnp.ndarray, epsilon: float
) -> jnp.ndarray:
  return jnp.exp(-remove_tensor_sum(cost_t, potentials) / epsilon)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial
from typing import Any, Dict, NamedTuple, Optional, Sequence, Tuple, Union

import jax
import jax.numpy as jnp

from ott import utils
from ott.geometry import pointcloud
from ott.math import fixed_point_loop
from ott.problems.linear import linear_problem
from ott.problems.quadratic import gw_barycenter
from ott.solvers import was_solver
from ott.solvers.quadratic import gromov_wasserstein

__all__ = ["GWBarycenterState", "GromovWassersteinBarycenter"]


class GWBarycenterState(NamedTuple):
  """State of the GW barycenter problem.

  Args:
    cost: Barycenter cost matrix of shape ``[bar_size, bar_size]``.
    x: Barycenter features of shape ``[bar_size, ndim_fused]``.
      Only used in the fused case.
    a: Weights of the barycenter of shape ``[bar_size,]``.
    errors: Array of shape
      ``[max_iter, num_measures, quad_max_iter, lin_outer_iter]`` containing
      the GW errors at each iteration.
    costs: Array of shape ``[max_iter,]`` containing the cost at each iteration.
    costs_bary: Array of shape ``[max_iter, num_measures]`` containing the
      cost between the individual measures and the barycenter at each iteration.
    gw_convergence: Array of shape ``[max_iter,]`` containing the convergence
      of all GW problems at each iteration.
  """
  cost: Optional[jnp.ndarray] = None
  x: Optional[jnp.ndarray] = None
  a: Optional[jnp.ndarray] = None
  errors: Optional[jnp.ndarray] = None
  costs: Optional[jnp.ndarray] = None
  costs_bary: Optional[jnp.ndarray] = None
  gw_convergence: Optional[jnp.ndarray] = None

  def set(self, **kwargs: Any) -> "GWBarycenterState":
    """Return a copy of self, possibly with overwrites."""
    return self._replace(**kwargs)

  @property
  def n_iters(self) -> int:
    """Number of iterations."""
    if self.gw_convergence is None:
      return -1
    return jnp.sum(self.gw_convergence != -1)


@jax.tree_util.register_pytree_node_class
class GromovWassersteinBarycenter(was_solver.WassersteinSolver):
  """Gromov-Wasserstein barycenter solver.

  Args:
    quadratic_solver: Quadratic OT solver.
    threshold: Convergence threshold.
    min_iterations: Minimum number of iterations.
    max_iterations: Maximum number of outermost iterations.
    store_inner_errors: Whether to store the errors of the quadratic OT solver.
  """

  def __init__(
      self,
      quadratic_solver: gromov_wasserstein.GromovWasserstein,
      threshold: float = 1e-3,
      min_iterations: int = 5,
      max_iterations: int = 50,
      store_inner_errors: bool = False,
  ):
    super().__init__(
        quadratic_solver.linear_solver,
        threshold=threshold,
        min_iterations=min_iterations,
        max_iterations=max_iterations,
        store_inner_errors=store_inner_errors,
    )
    self.quadratic_solver = quadratic_solver

  def __call__(
      self, problem: gw_barycenter.GWBarycenterProblem, bar_size: int,
      **kwargs: Any
  ) -> GWBarycenterState:
    """Solver the (fused) GW barycenter problem.

    Args:
      problem: The GW barycenter problem.
      bar_size: Size of the barycenter.
      kwargs: Keyword arguments for :meth:`init_state`.

    Returns:
      The solution.
    """
    state = self.init_state(problem, bar_size, **kwargs)
    state = iterations(self, problem, state)
    return self.output_from_state(state)

  def init_state(
      self,
      problem: gw_barycenter.GWBarycenterProblem,
      bar_size: int,
      bar_init: Optional[Union[jnp.ndarray, Tuple[jnp.ndarray,
                                                  jnp.ndarray]]] = None,
      a: Optional[jnp.ndarray] = None,
      rng: Optional[jax.Array] = None,
  ) -> GWBarycenterState:
    """Initialize the (fused) Gromov-Wasserstein barycenter state.

    Args:
      problem: The barycenter problem.
      bar_size: Size of the barycenter.
      bar_init: Initial barycenter value. Can be one of the following:

        - ``None`` - randomly initialize the barycenter.
        - :class:`jax.numpy.ndarray` - barycenter cost matrix of shape
          ``[bar_size, bar_size]``.
          Only used in the non-fused case.
        - :class:`tuple` of :class:`jax.numpy.ndarray` - the first array
          corresponds to a cost matrix of shape ``[bar_size, bar_size]``,
          the second array is a ``[bar_size, ndim_fused]`` feature matrix used
          in the fused case.

      a: An array of shape ``[bar_size,]`` containing the barycenter weights.
      rng: Random key for seeding used when ``bar_init = None``.

    Returns:
      The initial barycenter state.
    """
    if a is None:
      a = jnp.ones((bar_size,)) / bar_size
    else:
      assert a.shape == (bar_size,)

    if bar_init is None:
      rng = utils.default_prng_key(rng)
      _, b = problem.segmented_y_b
      rngs = jax.random.split(rng, problem.num_measures)

      transports = init_transports(
          self.linear_solver, rngs, a, b, problem.epsilon
      )
      x = problem.update_features(transports, a) if problem.is_fused else None
      cost = problem.update_barycenter(transports, a)
    else:
      cost, x = bar_init if isinstance(bar_init, tuple) else (bar_init, None)
      assert cost.shape == (bar_size, bar_size)
      if problem.is_fused:
        assert x is not None, "Barycenter features are not initialized."
        assert x.shape == (bar_size, problem.ndim_fused)

    num_iter = self.max_iterations
    if self.store_inner_errors:
      # TODO(michalk8): in the future, think about how to do this in general
      errors = -jnp.ones((
          num_iter, problem.num_measures, self.quadratic_solver.max_iterations,
          self.linear_solver.outer_iterations
      ))
    else:
      errors = None

    costs = -jnp.ones((num_iter,))
    costs_bary = -jnp.ones((num_iter, problem.num_measures))
    gw_convergence = -jnp.ones((num_iter,))
    return GWBarycenterState(
        cost=cost,
        x=x,
        a=a,
        errors=errors,
        costs=costs,
        costs_bary=costs_bary,
        gw_convergence=gw_convergence
    )

  def update_state(
      self,
      state: GWBarycenterState,
      iteration: int,
      problem: gw_barycenter.GWBarycenterProblem,
      store_errors: bool = True,
  ) -> Tuple[float, bool, jnp.ndarray, Optional[jnp.ndarray]]:
    """Solve the (fused) Gromov-Wasserstein barycenter problem."""

    def solve_gw(
        state: GWBarycenterState, b: jnp.ndarray, y: jnp.ndarray,
        f: Optional[jnp.ndarray]
    ) -> Tuple[float, bool, jnp.ndarray, Optional[jnp.ndarray]]:
      quad_problem = problem._create_problem(state, y=y, b=b, f=f)
      out = self.quadratic_solver(quad_problem)
      return (
          out.reg_gw_cost, out.converged, out.matrix,
          out.errors if store_errors else None
      )

    in_axes = [None, 0, 0]
    in_axes += [0] if problem.is_fused else [None]
    solve_fn = jax.vmap(solve_gw, in_axes=in_axes)

    y, b = problem.segmented_y_b
    y_f = problem.segmented_y_fused
    costs, convergeds, transports, errors = solve_fn(state, b, y, y_f)

    cost = jnp.sum(costs * problem.weights)
    costs_bary = state.costs_bary.at[iteration].set(costs)
    costs = state.costs.at[iteration].set(cost)

    converged = jnp.all(convergeds)
    gw_convergence = state.gw_convergence.at[iteration].set(converged)

    if self.store_inner_errors:
      errors = state.errors.at[iteration, ...].set(errors)
    else:
      errors = None

    x = problem.update_features(
        transports, state.a
    ) if problem.is_fused else state.x
    cost = problem.update_barycenter(transports, state.a)
    return state.set(
        cost=cost,
        x=x,
        costs=costs,
        costs_bary=costs_bary,
        errors=errors,
        gw_convergence=gw_convergence
    )

  def output_from_state(self, state: GWBarycenterState) -> GWBarycenterState:
    """No-op."""
    # TODO(michalk8): just for consistency with continuous barycenter
    # will be refactored in the future to create an output
    return state

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:  # noqa: D102
    return ([self.quadratic_solver, self.threshold], {
        "min_iterations": self.min_iterations,
        "max_iterations": self.max_iterations,
        "store_inner_errors": self.store_inner_errors,
    })


@partial(jax.vmap, in_axes=[None, 0, None, 0, None])
def init_transports(
    solver, rng: jax.Array, a: jnp.ndarray, b: jnp.ndarray,
    epsilon: Optional[float]
) -> jnp.ndarray:
  """Initialize random 2D point cloud and solve the linear OT problem.

  Args:
    solver: Linear OT solver.
    rng: Random key for seeding.
    a: Source marginals (e.g., for barycenter) of shape ``[bar_size,]``.
    b: Target marginals of shape ``[max_measure_size,]``.
    epsilon: Entropy regularization.

  Returns:
    Transport map of shape ``[bar_size, max_measure_size]``.
  """
  rng1, rng2 = jax.random.split(rng, 2)
  x = jax.random.normal(rng1, shape=(len(a), 2))
  y = jax.random.normal(rng2, shape=(len(b), 2))
  geom = pointcloud.PointCloud(
      x, y, epsilon=epsilon, src_mask=a > 0, tgt_mask=b > 0
  )
  problem = linear_problem.LinearProblem(geom, a=a, b=b)
  return solver(problem).matrix


def iterations(  # noqa: D103
    solver: GromovWassersteinBarycenter,
    problem: gw_barycenter.GWBarycenterProblem, init_state: GWBarycenterState
) -> GWBarycenterState:

  def cond_fn(
      iteration: int, constants: GromovWassersteinBarycenter,
      state: GWBarycenterState
  ) -> bool:
    solver, _ = constants
    return solver._continue(state, iteration)

  def body_fn(
      iteration, constants: Tuple[GromovWassersteinBarycenter,
                                  gw_barycenter.GWBarycenterProblem],
      state: GWBarycenterState, compute_error: bool
  ) -> GWBarycenterState:
    del compute_error  # always assumed true
    solver, problem = constants
    return solver.update_state(state, iteration, problem)

  return fixed_point_loop.fixpoint_iter(
      cond_fn=cond_fn,
      body_fn=body_fn,
      min_iterations=solver.min_iterations,
      max_iterations=solver.max_iterations,
      inner_iterations=1,
      constants=(solver, problem),
      state=init_state,
  )



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import Any, NamedTuple, Optional, Tuple

import jax
import jax.numpy as jnp

from ott import utils
from ott.geometry import pointcloud
from ott.math import fixed_point_loop
from ott.math import utils as mu
from ott.problems.linear import barycenter_problem, linear_problem
from ott.solvers import was_solver

__all__ = ["FreeBarycenterState", "FreeWassersteinBarycenter"]


class FreeBarycenterState(NamedTuple):
  """Holds the state of the Wasserstein barycenter solver.

  Args:
    costs: Holds the sequence of regularized GW costs seen through the outer
      loop of the solver.
    linear_convergence: Holds the sequence of bool convergence flags of the
      inner Sinkhorn iterations.
    errors: Holds sequence of vectors of errors of the Sinkhorn algorithm
      at each iteration.
    x: barycenter points.
    a: barycenter weights.
  """

  costs: Optional[jnp.ndarray] = None
  linear_convergence: Optional[jnp.ndarray] = None
  errors: Optional[jnp.ndarray] = None
  x: Optional[jnp.ndarray] = None
  a: Optional[jnp.ndarray] = None

  def set(self, **kwargs: Any) -> "FreeBarycenterState":
    """Return a copy of self, possibly with overwrites."""
    return self._replace(**kwargs)

  def update(
      self, iteration: int, bar_prob: barycenter_problem.FreeBarycenterProblem,
      linear_solver: Any, store_errors: bool
  ) -> "FreeBarycenterState":
    """Update the state of the solver.

    Args:
      iteration: the current iteration of the outer loop.
      bar_prob: the barycenter problem.
      linear_solver: the linear OT solver to use.
      store_errors: whether to store the errors of the inner loop.

    Returns:
      The updated state.
    """
    seg_y, seg_b = bar_prob.segmented_y_b

    @functools.partial(jax.vmap, in_axes=[None, None, 0, 0])
    def solve_linear_ot(
        a: Optional[jnp.ndarray], x: jnp.ndarray, b: jnp.ndarray, y: jnp.ndarray
    ):
      geom = pointcloud.PointCloud(
          x,
          y,
          src_mask=a > 0.0,
          tgt_mask=b > 0.0,
          cost_fn=bar_prob.cost_fn,
          epsilon=bar_prob.epsilon
      )
      prob = linear_problem.LinearProblem(geom, a=a, b=b)
      out = linear_solver(prob)
      return (
          out.reg_ot_cost, out.converged, out.matrix,
          out.errors if store_errors else None
      )

    reg_ot_costs, convergeds, matrices, errors = solve_linear_ot(
        self.a, self.x, seg_b, seg_y
    )

    cost = jnp.sum(reg_ot_costs * bar_prob.weights)
    updated_costs = self.costs.at[iteration].set(cost)
    converged = jnp.all(convergeds)
    linear_convergence = self.linear_convergence.at[iteration].set(converged)

    if store_errors and self.errors is not None:
      errors = self.errors.at[iteration, :, :].set(errors)
    else:
      errors = None

    # Approximation of barycenter as barycenter of barycenters per measure.

    barycenters_per_measure = mu.barycentric_projection(
        matrices, seg_y, bar_prob.cost_fn
    )

    x_new = jax.vmap(
        lambda w, y: bar_prob.cost_fn.barycenter(w, y)[0], in_axes=[None, 1]
    )(bar_prob.weights, barycenters_per_measure)

    return self.set(
        costs=updated_costs,
        linear_convergence=linear_convergence,
        errors=errors,
        x=x_new
    )


@jax.tree_util.register_pytree_node_class
class FreeWassersteinBarycenter(was_solver.WassersteinSolver):
  """Continuous Wasserstein barycenter solver :cite:`cuturi:14`."""

  def __call__(  # noqa: D102
      self,
      bar_prob: barycenter_problem.FreeBarycenterProblem,
      bar_size: int = 100,
      x_init: Optional[jnp.ndarray] = None,
      rng: Optional[jax.Array] = None,
  ) -> FreeBarycenterState:
    # TODO(michalk8): no reason for iterations to be outside this class
    rng = utils.default_prng_key(rng)
    return iterations(self, bar_size, bar_prob, x_init, rng)

  def init_state(
      self,
      bar_prob: barycenter_problem.FreeBarycenterProblem,
      bar_size: int,
      x_init: Optional[jnp.ndarray] = None,
      rng: Optional[jax.Array] = None,
  ) -> FreeBarycenterState:
    """Initialize the state of the Wasserstein barycenter iterations.

    Args:
      bar_prob: The barycenter problem.
      bar_size: Size of the barycenter.
      x_init: Initial barycenter estimate of shape ``[bar_size, ndim]``.
        If `None`, ``bar_size`` points will be sampled from the input
        measures according to their weights
        :attr:`~ott.problems.linear.barycenter_problem.FreeBarycenterProblem.flattened_y`.
      rng: Random key for seeding.

    Returns:
      The initial barycenter state.
    """
    if x_init is not None:
      assert bar_size == x_init.shape[0]
      x = x_init
    else:
      # sample randomly points in the support of the y measures
      rng = utils.default_prng_key(rng)
      indices_subset = jax.random.choice(
          rng,
          a=bar_prob.flattened_y.shape[0],
          shape=(bar_size,),
          replace=False,
          p=bar_prob.flattened_b
      )
      x = bar_prob.flattened_y[indices_subset, :]

    # TODO(cuturi) expand to non-uniform weights for barycenter.
    a = jnp.ones((bar_size,)) / bar_size
    num_iter = self.max_iterations
    if self.store_inner_errors:
      errors = -jnp.ones((
          num_iter, bar_prob.num_measures, self.linear_solver.outer_iterations
      ))
    else:
      errors = None
    return FreeBarycenterState(
        -jnp.ones((num_iter,)), -jnp.ones((num_iter,)), errors, x, a
    )

  def output_from_state(  # noqa: D102
      self, state: FreeBarycenterState
  ) -> FreeBarycenterState:
    # TODO(michalk8): create an output variable to match rest of the framework
    return state


def iterations(
    solver: FreeWassersteinBarycenter, bar_size: int,
    bar_prob: barycenter_problem.FreeBarycenterProblem, x_init: jnp.ndarray,
    rng: jax.Array
) -> FreeBarycenterState:
  """Jittable Wasserstein barycenter outer loop."""

  def cond_fn(
      iteration: int,
      constants: Tuple[FreeWassersteinBarycenter,
                       barycenter_problem.FreeBarycenterProblem],
      state: FreeBarycenterState
  ) -> bool:
    solver, _ = constants
    return solver._continue(state, iteration)

  def body_fn(
      iteration, constants: Tuple[FreeWassersteinBarycenter,
                                  barycenter_problem.FreeBarycenterProblem],
      state: FreeBarycenterState, compute_error: bool
  ) -> FreeBarycenterState:
    del compute_error  # Always assumed True
    solver, bar_prob = constants
    return state.update(
        iteration, bar_prob, solver.linear_solver, solver.store_inner_errors
    )

  state = fixed_point_loop.fixpoint_iter(
      cond_fn=cond_fn,
      body_fn=body_fn,
      min_iterations=solver.min_iterations,
      max_iterations=solver.max_iterations,
      inner_iterations=1,
      constants=(solver, bar_prob),
      state=solver.init_state(bar_prob, bar_size, x_init, rng)
  )

  return solver.output_from_state(state)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable, Optional, TypeVar

import equinox as eqx
import lineax as lx
from jaxtyping import Array, Float, PyTree

import jax
import jax.numpy as jnp
import jax.tree_util as jtu

_T = TypeVar("_T")
_FlatPyTree = tuple[list[_T], jtu.PyTreeDef]

__all__ = ["CustomTransposeLinearOperator", "solve_lineax"]


class CustomTransposeLinearOperator(lx.FunctionLinearOperator):
  """Implement a linear operator that can specify its transpose directly."""
  fn: Callable[[PyTree[Float[Array, "..."]]], PyTree[Float[Array, "..."]]]
  fn_t: Callable[[PyTree[Float[Array, "..."]]], PyTree[Float[Array, "..."]]]
  input_structure: _FlatPyTree[jax.ShapeDtypeStruct] = eqx.static_field()
  input_structure_t: _FlatPyTree[jax.ShapeDtypeStruct] = eqx.static_field()
  tags: frozenset[object]

  def __init__(self, fn, fn_t, input_structure, input_structure_t, tags=()):
    super().__init__(fn, input_structure, tags)
    self.fn_t = eqx.filter_closure_convert(fn_t, input_structure_t)
    self.input_structure_t = input_structure_t

  def transpose(self):
    """Provide custom transposition operator from function."""
    return lx.FunctionLinearOperator(self.fn_t, self.input_structure_t)


def solve_lineax(
    lin: Callable,
    b: jnp.ndarray,
    lin_t: Optional[Callable] = None,
    symmetric: bool = False,
    nonsym_solver: Optional[lx.AbstractLinearSolver] = None,
    ridge_identity: float = 0.0,
    ridge_kernel: float = 0.0,
    **kwargs: Any
) -> jnp.ndarray:
  """Wrapper around lineax solvers.

  Args:
    lin: Linear operator
    b: vector. Returned `x` is such that `lin(x)=b`
    lin_t: Linear operator, corresponding to transpose of `lin`.
    symmetric: whether `lin` is symmetric.
    nonsym_solver: solver used when handling non-symmetric cases. Note that
      :class:`~lineax.CG` is used by default in the symmetric case.
    ridge_kernel: promotes zero-sum solutions. Only use if `tau_a = tau_b = 1.0`
    ridge_identity: handles rank deficient transport matrices (this happens
      typically when rows/cols in cost/kernel matrices are collinear, or,
      equivalently when two points from either measure are close).
    kwargs: arguments passed to :class:`~lineax.AbstractLinearSolver` linear
      solver.
  """
  input_structure = jax.eval_shape(lambda: b)
  kwargs.setdefault("rtol", 1e-6)
  kwargs.setdefault("atol", 1e-6)

  if ridge_kernel > 0.0 or ridge_identity > 0.0:
    lin_reg = lambda x: lin(x) + ridge_kernel * jnp.sum(x) + ridge_identity * x
    lin_t_reg = lambda x: lin_t(x) + ridge_kernel * jnp.sum(
        x
    ) + ridge_identity * x
  else:
    lin_reg, lin_t_reg = lin, lin_t

  if symmetric:
    solver = lx.CG(**kwargs)
    fn_operator = lx.FunctionLinearOperator(
        lin_reg, input_structure, tags=lx.positive_semidefinite_tag
    )
    return lx.linear_solve(fn_operator, b, solver).value
  # In the non-symmetric case, use NormalCG by default, but consider
  # user defined choice of alternative lx solver.
  solver_type = lx.NormalCG if nonsym_solver is None else nonsym_solver
  solver = solver_type(**kwargs)
  fn_operator = CustomTransposeLinearOperator(
      lin_reg, lin_t_reg, input_structure, input_structure
  )
  return lx.linear_solve(fn_operator, b, solver).value



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import TYPE_CHECKING, Optional, Sequence, Tuple, Union

import jax
import jax.numpy as jnp
import jax.scipy as jsp

if TYPE_CHECKING:
  from ott.geometry import costs

__all__ = [
    "safe_log",
    "norm",
    "kl",
    "gen_kl",
    "gen_js",
    "gen_ent",
    "logsumexp",
    "softmin",
    "barycentric_projection",
    "sort_and_argsort",
    "lambertw",
]


def safe_log(  # noqa: D103
    x: jnp.ndarray,
    *,
    eps: Optional[float] = None
) -> jnp.ndarray:
  if eps is None:
    eps = jnp.finfo(x.dtype).tiny
  return jnp.where(x > 0.0, jnp.log(x), jnp.log(eps))


@functools.partial(jax.custom_jvp, nondiff_argnums=[1, 2, 3])
@functools.partial(jax.jit, static_argnames=("ord", "axis", "keepdims"))
def norm(
    x: jnp.ndarray,
    ord: Union[int, str, None] = None,
    axis: Union[None, Sequence[int], int] = None,
    keepdims: bool = False
) -> jnp.ndarray:
  """Computes order ord norm of vector, using `jnp.linalg` in forward pass.

  Evaluations of distances between a vector and itself using translation
  invariant costs, typically  norms, result in functions of the form
  ``lambda x : jnp.linalg.norm(x-x)``. Such functions output `NaN` gradients,
  because they involve computing the derivative of a negative exponent of 0
  (e.g. when differentiating the Euclidean norm, one gets a 0-denominator in the
  expression, see e.g. https://github.com/jax-ml/jax/issues/6484 for context).

  While this makes sense mathematically, in the context of optimal transport
  such distances between a point and itself can be safely ignored when they
  contribute to an OT cost (when, for instance, computing Sinkhorn divergences,
  involving computing the OT cost of a point cloud with itself).

  To avoid such `NaN` values, this custom norm implementation uses the
  double-where trick, to avoid having branches that output any `NaN`, and
  safely output a 0 instead.

  Args:
    x: Input array.  If `axis` is None, `x` must be 1-D or 2-D, unless `ord`
      is None. If both `axis` and `ord` are None, the 2-norm of ``x.ravel``
      will be returned.
    ord: `{non-zero int, jnp.inf, -jnp.inf, 'fro', 'nuc'}`, Order of the norm.
      The default is `None`, which is equivalent to `2.0` for vectors.
    axis: `{None, int, 2-tuple of ints}`, optional. If `axis` is an integer, it
      specifies the axis of `x` along which to compute the vector norms.
      If `axis` is a 2-tuple, it specifies the axes that hold 2-D matrices, and
      the matrix norms of these matrices are computed.  If `axis` is None then
      either a vector norm (when `x` is 1-D) or a matrix norm (when `x` is 2-D)
      is returned. The default is None.
    keepdims: If set to True, the axes which are normed over are left in the
      result as dimensions with size one.  With this option the result will
      broadcast correctly against the original `x`.

  Returns:
    float or ndarray, Norm of the matrix or vector(s).
  """
  return jnp.linalg.norm(x, ord=ord, axis=axis, keepdims=keepdims)


@norm.defjvp
def norm_jvp(ord, axis, keepdims, primals, tangents):
  """Custom_jvp for norm, that returns 0.0 when evaluated at 0."""
  x, = primals
  x_is_zero = jnp.all(jnp.logical_not(x))
  clean_x = jnp.where(x_is_zero, jnp.ones_like(x), x)
  primals, tangents = jax.jvp(
      functools.partial(jnp.linalg.norm, ord=ord, axis=axis, keepdims=keepdims),
      (clean_x,), tangents
  )
  return primals, jnp.where(x_is_zero, 0.0, tangents)


# TODO(michalk8): add axis argument
def kl(p: jnp.ndarray, q: jnp.ndarray) -> float:
  """Kullback-Leibler divergence."""
  return jnp.vdot(p, (safe_log(p) - safe_log(q)))


def gen_ent(x: jnp.ndarray) -> float:
  """Generalized entropy, adds the sum of ``x`` compared to usual entropy."""
  return jnp.sum(jsp.special.entr(x)) + jnp.sum(x)


def gen_kl(p: jnp.ndarray, q: jnp.ndarray) -> float:
  """Generalized Kullback-Leibler divergence."""
  return jnp.vdot(p, (safe_log(p) - safe_log(q))) + jnp.sum(q) - jnp.sum(p)


# TODO(michalk8): add axis argument
def gen_js(p: jnp.ndarray, q: jnp.ndarray, c: float = 0.5) -> float:
  """Jensen-Shannon divergence."""
  return c * (gen_kl(p, q) + gen_kl(q, p))


@functools.partial(jax.custom_jvp, nondiff_argnums=(1, 2, 4))
def logsumexp(  # noqa: D103
    mat, axis=None, keepdims=False, b=None, return_sign=False
):
  return jax.scipy.special.logsumexp(
      mat, axis=axis, keepdims=keepdims, b=b, return_sign=return_sign
  )


@logsumexp.defjvp
def logsumexp_jvp(axis, keepdims, return_sign, primals, tangents):
  """Custom derivative rule for lse that does not blow up with -inf.

  This logsumexp implementation uses the standard jax one in forward mode but
  implements a custom rule to differentiate. Given the preference of jax for
  jvp over vjp, and the fact that this is a simple linear rule, jvp is used.
  This custom differentiation address issues when the output of lse is
  -inf (which corresponds to the case where all inputs in a slice are -inf,
  which happens typically when ``a`` or ``b`` weight vectors have zeros.)

  Although both exp(lse) and its derivative should be 0, automatic
  differentiation returns a NaN derivative because of a -inf - (-inf) operation
  appearing in the definition of centered_exp below. This is corrected in the
  implementation below.

  Args:
    axis: argument from original logsumexp
    keepdims: argument from original logsumexp
    return_sign: argument from original logsumexp
    primals: mat and b, the two arguments against which we differentiate.
    tangents: of same size as mat and b.

  Returns:
    original primal outputs + their tangent.
  """  # noqa: D401
  mat, b = primals
  tan_mat, tan_b = tangents
  lse = logsumexp(mat, axis, keepdims, b, return_sign)
  if return_sign:
    lse, sign = lse
  lse = jnp.where(jnp.isfinite(lse), lse, 0.0)

  if axis is not None:
    centered_exp = jnp.exp(mat - jnp.expand_dims(lse, axis=axis))
  else:
    centered_exp = jnp.exp(mat - lse)

  if b is None:
    res = jnp.sum(centered_exp * tan_mat, axis=axis, keepdims=keepdims)
  else:
    res = jnp.sum(b * centered_exp * tan_mat, axis=axis, keepdims=keepdims)
    res += jnp.sum(tan_b * centered_exp, axis=axis, keepdims=keepdims)
  if return_sign:
    return (lse, sign), (sign * res, jnp.zeros_like(sign))
  return lse, res


@functools.partial(jax.custom_vjp, nondiff_argnums=(2,))
def softmin(
    x: jnp.ndarray,
    gamma: float,
    axis: Optional[Union[int, Sequence[int]]] = None
) -> jnp.ndarray:
  r"""Soft-min operator.

  Args:
    x: Input data.
    gamma: Smoothing parameter :math:`> 0`.
    axis: Axis or axes over which to operate. If ``None``, use flattened input.

  Returns:
    The soft minimum.
  """
  return -gamma * jsp.special.logsumexp(x / -gamma, axis=axis)


softmin.defvjp(
    lambda x, gamma, axis: (softmin(x, gamma, axis), (x / -gamma, axis)),
    lambda axis, res, g: (
        jnp.where(
            jnp.isinf(res[0]), 0.0,
            jax.nn.softmax(res[0], axis=axis) *
            (g if axis is None else jnp.expand_dims(g, axis=axis))
        ), None
    )
)


@functools.partial(jax.vmap, in_axes=[0, 0, None])
def barycentric_projection(
    matrix: jnp.ndarray, y: jnp.ndarray, cost_fn: "costs.CostFn"
) -> jnp.ndarray:
  """Compute the barycentric projection of a matrix.

  Args:
    matrix: a matrix of shape (n, m)
    y: a vector of shape (m,)
    cost_fn: a CostFn instance.

  Returns:
    a vector of shape (n,) containing the barycentric projection of matrix.
  """
  return jax.vmap(
      lambda m, y: cost_fn.barycenter(m, y)[0], in_axes=[0, None]
  )(matrix, y)


def sort_and_argsort(
    x: jnp.array,
    *,
    argsort: bool = False
) -> Tuple[jnp.ndarray, Optional[jnp.ndarray]]:
  """Unified function that returns both sort and argsort, if latter needed."""
  if argsort:
    i_x = jnp.argsort(x)
    return x[i_x], i_x
  return jnp.sort(x), None


@functools.partial(jax.custom_jvp, nondiff_argnums=(1, 2))
def lambertw(
    z: jnp.ndarray, tol: float = 1e-8, max_iter: int = 100
) -> jnp.ndarray:
  """Principal branch of the
  `Lambert W function <https://en.wikipedia.org/wiki/Lambert_W_function>`_.

  This implementation uses Halley's iteration and the global initialization
  proposed in :cite:`iacono:17`, Eq. 20 .

  Args:
     z: Array.
     tol: Tolerance threshold.
     max_iter: Maximum number of iterations.

  Returns:
    The Lambert W evaluated at ``z``.
  """  # noqa: D205

  def initial_iacono(x: jnp.ndarray) -> jnp.ndarray:
    y = jnp.sqrt(1.0 + jnp.e * x)
    num = 1.0 + 1.14956131 * y
    denom = 1.0 + 0.45495740 * jnp.log1p(y)
    return -1.0 + 2.036 * jnp.log(num / denom)

  def cond_fun(container):
    it, converged, _ = container
    return jnp.logical_and(jnp.any(~converged), it < max_iter)

  def halley_iteration(container):
    it, _, w = container

    # modified from `tensorflow_probability`
    f = w - z * jnp.exp(-w)
    delta = f / (w + 1.0 - 0.5 * (w + 2.0) * f / (w + 1.0))

    w_next = w - delta

    not_converged = jnp.abs(delta) <= tol * jnp.abs(w_next)
    return it + 1, not_converged, w_next

  w0 = initial_iacono(z)
  converged = jnp.zeros_like(w0, dtype=bool)

  _, _, w = jax.lax.while_loop(
      cond_fun=cond_fun, body_fun=halley_iteration, init_val=(0, converged, w0)
  )
  return w


@lambertw.defjvp
def _lambertw_jvp(
    tol: float, max_iter: int, primals: Tuple[jnp.ndarray, ...],
    tangents: Tuple[jnp.ndarray, ...]
) -> Tuple[jnp.ndarray, jnp.ndarray]:
  z, = primals
  dz, = tangents
  w = lambertw(z, tol=tol, max_iter=max_iter)
  pz = jnp.where(z == 0.0, 1.0, w / ((1.0 + w) * z))
  return w, pz * dz



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import Any, Dict, Literal, Optional, Sequence, Tuple, Union

import jax
import jax.numpy as jnp

from ott.geometry import costs, geometry, pointcloud, segment
from ott.math import utils as mu
from ott.problems.linear import barycenter_problem
from ott.problems.quadratic import quadratic_costs, quadratic_problem

__all__ = ["GWBarycenterProblem"]


# TODO(michalk8): better abstraction (common superclass for Wasserstein bary)
@jax.tree_util.register_pytree_node_class
class GWBarycenterProblem(barycenter_problem.FreeBarycenterProblem):
  """(Fused) Gromov-Wasserstein barycenter problem :cite:`peyre:16,vayer:19`.

  Args:
    y: Array of shape ``[num_total_points, ndim]`` merging the points of all
      measures. Alternatively, already segmented array of shape
      ``[num_measures, max_measure_size, ndim]`` can be passed.
      See also :func:`~ott.geometry.segment.segment_point_cloud`.
    b: Array of shape ``[num_total_points,]`` containing the weights of all
      the points within the measures that define the barycenter problem.
      Same as ``y``, pre-segmented array of weights of shape
      ``[num_measures, max_measure_size]`` can be passed.
      If ``y`` is already pre-segmented, this array must be passed.
    weights: Array of shape ``[num_measures,]`` containing the weights of the
      barycenter problem.
    costs: Alternative to ``y``, an array of shape
      ``[num_measures, max_measure_size, max_measure_size]`` that defines padded
      cost matrices for each measure. Used in the quadratic term.
      Only one of ``y`` and ``cost`` can be specified.
    y_fused: Array of shape ``[num_total_points, ndim_fused]`` containing
      the data of the points of all measures used to define the linear term
      in the fused case. Same as ``y``, it can be specified as a pre-segmented
      array of shape ``[num_measures, max_measure_size, ndim_fused]``.
    gw_loss: Gromov-Wasserstein loss.
    fused_penalty: Multiplier of the linear term. Only used when
      ``y_fused != None``.
    scale_cost: Scaling of cost matrices passed to geometries.
    kwargs: Keyword arguments for
      :class:`~ott.problems.linear.barycenter_problem.BarycenterProblem`.
  """

  def __init__(
      self,
      y: Optional[jnp.ndarray] = None,
      b: Optional[jnp.ndarray] = None,
      weights: Optional[jnp.ndarray] = None,
      costs: Optional[jnp.ndarray] = None,
      y_fused: Optional[jnp.ndarray] = None,
      fused_penalty: float = 1.0,
      gw_loss: Literal["sqeucl", "kl"] = "sqeucl",
      scale_cost: Union[float, Literal["mean", "max_cost"]] = 1.0,
      **kwargs: Any,
  ):
    assert y is None or costs is None, "Cannot specify both `y` and `costs`."
    y = y if costs is None else costs

    super().__init__(y=y, b=b, weights=weights, **kwargs)

    self._y_fused = y_fused
    self.fused_penalty = fused_penalty
    self._loss_name = gw_loss
    self.scale_cost = scale_cost
    self._y_as_costs = costs is not None

    if self._y_as_costs:
      # (num_measures, max_measure_size, max_measure_size)
      _, n, m = self._y.shape
      assert n == m, "Cost matrices must be square."
    if self.is_fused:
      seg_y = self._is_segmented
      seg_fused = self._y_fused.ndim == 3
      if seg_y and seg_fused:
        # (num_measures, max_measure_size, ndim_fused)
        # (num_measures, max_measure_size, ndim)
        assert self._y_fused.shape[:2] == self._y.shape[:2]
      if not seg_y and not seg_fused:
        # (num_total_points, ndim_fused), (num_total_points, ndim)
        assert self._y_fused.shape[0] == self._y.shape[0]
      # TODO(michalk8): in the future, consider checking the other 2 cases
      # using `segmented_y` and `segmented_y_fused`?

  def update_barycenter(
      self, transports: jnp.ndarray, a: jnp.ndarray
  ) -> jnp.ndarray:
    """Update the barycenter cost matrix.

    Uses the eq. 14 and 15 of :cite:`peyre:16`.

    Args:
      transports: Transport maps of shape
        ``[num_measures, bar_size, max_measure_size]``.
      a: Barycenter weights of shape ``[bar_size,]``.

    Returns:
      Update cost matrix of shape ``[bar_size, bar_size]``.
    """

    @functools.partial(jax.vmap, in_axes=[0, 0, 0, None])
    def project(
        y: jnp.ndarray,
        b: jnp.ndarray,
        transport: jnp.ndarray,
        fn: Optional[quadratic_costs.Loss],
    ) -> jnp.ndarray:
      geom = self._create_y_geometry(y, mask=b > 0.0)
      fn, lin = (None, True) if fn is None else (fn.func, fn.is_linear)

      tmp = geom.apply_cost(
          transport.T,
          axis=0,
          fn=fn,
          is_linear=lin,
      )
      return transport @ tmp

    fn = None if self._loss_name == "sqeucl" else self.gw_loss.h2
    y, b = self.segmented_y_b
    weights = self.weights[:, None, None]

    barycenter = jnp.sum(weights * project(y, b, transports, fn), axis=0)
    inv_a = jnp.where(a > 0, 1.0 / a, 1.0)
    barycenter = (barycenter * inv_a[None, :]) * inv_a[:, None]

    # TODO(michalk8): in future, use `isinstanceof(self.gw_loss, ...)`
    # once refactoring has been done
    if self._loss_name == "kl":
      return jnp.exp(barycenter)
    return barycenter

  def update_features(self, transports: jnp.ndarray,
                      a: jnp.ndarray) -> Optional[jnp.ndarray]:
    """Update the barycenter features in the fused case :cite:`vayer:19`.

    Uses :cite:`cuturi:14` eq. 8, and is implemented only
    for the :class:`~ott.geometry.costs.SqEuclidean` cost.

    Args:
      transports: Transport maps of shape
        ``[num_measures, bar_size, max_measure_size]``.
      a: Barycenter weights of shape ``[bar_size,]``.

    Returns:
      Updated features of shape ``[bar_size, ndim_fused]``.
    """
    if not self.is_fused:
      raise RuntimeError(
          "Updating features is available only in the fused case."
      )

    y_fused = self.segmented_y_fused
    weights = self.weights[:, None, None]
    inv_a = jnp.where(a > 0, 1.0 / a, 1.0)
    transports = transports * inv_a[None, :, None]

    if self._loss_name == "sqeucl":
      cost_fn = costs.SqEuclidean()
      return jnp.sum(
          weights * mu.barycentric_projection(transports, y_fused, cost_fn),
          axis=0
      )
    raise NotImplementedError(self._loss_name)

  def _create_bary_geometry(
      self,
      cost_matrix: jnp.ndarray,
      mask: Optional[jnp.ndarray] = None
  ) -> geometry.Geometry:
    return geometry.Geometry(
        cost_matrix=cost_matrix,
        src_mask=mask,
        tgt_mask=mask,
        epsilon=self.epsilon,
        scale_cost=self.scale_cost
    )

  def _create_y_geometry(
      self,
      y: jnp.ndarray,
      mask: Optional[jnp.ndarray] = None
  ) -> geometry.Geometry:
    if self._y_as_costs:
      assert y.shape[0] == y.shape[1], y.shape
      return geometry.Geometry(
          y,
          epsilon=self.epsilon,
          scale_cost=self.scale_cost,
          src_mask=mask,
          tgt_mask=mask
      )
    return pointcloud.PointCloud(
        y,
        epsilon=self.epsilon,
        scale_cost=self.scale_cost,
        cost_fn=self.cost_fn,
        src_mask=mask,
        tgt_mask=mask
    )

  def _create_fused_geometry(
      self,
      x: jnp.ndarray,
      y: jnp.ndarray,
      src_mask: Optional[jnp.ndarray] = None,
      tgt_mask: Optional[jnp.ndarray] = None
  ) -> pointcloud.PointCloud:
    return pointcloud.PointCloud(
        x,
        y,
        cost_fn=self.cost_fn,
        epsilon=self.epsilon,
        scale_cost=self.scale_cost,
        src_mask=src_mask,
        tgt_mask=tgt_mask
    )

  def _create_problem(
      self,
      state: "GWBarycenterState",  # noqa: F821
      y: jnp.ndarray,
      b: jnp.ndarray,
      f: Optional[jnp.ndarray] = None
  ) -> quadratic_problem.QuadraticProblem:
    # TODO(michalk8): in future, mask in the problem for convenience?
    bary_mask = state.a > 0.0
    y_mask = b > 0.0

    geom_xx = self._create_bary_geometry(state.cost, mask=bary_mask)
    geom_yy = self._create_y_geometry(y, mask=y_mask)
    if self.is_fused:
      assert f is not None
      assert state.x.shape[1] == f.shape[1]
      geom_xy = self._create_fused_geometry(
          state.x, f, src_mask=bary_mask, tgt_mask=y_mask
      )
    else:
      geom_xy = None

    return quadratic_problem.QuadraticProblem(
        geom_xx=geom_xx,
        geom_yy=geom_yy,
        geom_xy=geom_xy,
        a=state.a,
        b=b,
        fused_penalty=self.fused_penalty,
    )

  @property
  def is_fused(self) -> bool:
    """Whether the problem is fused."""
    return self._y_fused is not None

  @property
  def segmented_y_fused(self) -> Optional[jnp.ndarray]:
    """Feature array of shape used in the fused case."""
    if not self.is_fused or self._y_fused.ndim == 3:
      return self._y_fused
    y_fused, _ = segment.segment_point_cloud(
        x=self._y_fused,
        padding_vector=self.cost_fn._padder(self.ndim_fused),
        **self._kwargs
    )
    return y_fused

  @property
  def ndim(self) -> Optional[int]:  # noqa: D102
    return None if self._y_as_costs else self._y.shape[-1]

  @property
  def ndim_fused(self) -> Optional[int]:
    """Number of dimensions of the fused term."""
    return self._y_fused.shape[-1] if self.is_fused else None

  @property
  def gw_loss(self) -> quadratic_costs.GWLoss:
    """Gromov-Wasserstein loss."""
    # TODO(michalk8): custom losses would require inverting some fns;
    # `https://jax.readthedocs.io/en/latest/notebooks/ some fns;
    # Writing_custom_interpreters_in_Jax.html#your-first-interpreter-invert`
    # might be useful
    if self._loss_name == "sqeucl":
      return quadratic_costs.make_square_loss()
    if self._loss_name == "kl":
      return quadratic_costs.make_kl_loss()
    raise NotImplementedError(
        f"Loss `{self._loss_name}` is not yet implemented."
    )

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:  # noqa: D102
    (y, b, weights), aux = super().tree_flatten()
    if self._y_as_costs:
      children = [None, b, weights, y]
    else:
      children = [y, b, weights, None]
    aux["fused_penalty"] = self.fused_penalty
    aux["gw_loss"] = self._loss_name
    aux["scale_cost"] = self.scale_cost
    return children + [self._y_fused], aux

  @classmethod
  def tree_unflatten(  # noqa: D102
      cls, aux_data: Dict[str, Any], children: Sequence[Any]
  ) -> "GWBarycenterProblem":
    y, b, weights, costs, y_fused = children
    return cls(
        y=y, b=b, weights=weights, costs=costs, y_fused=y_fused, **aux_data
    )
